from noora.io.Filterable import Filterable


class FileFolderFilter(Filterable):
    def __init__(self, file):
        Filterable.__init__(self)
        self.__file = file

    def get_file(self):
        return self.__file

    def accept(self, fileable):
        file = self.get_file()
        if fileable.is_dir() and fileable.tail() == file.tail():
            return True
        return False



import os
from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader
from noora.exceptions.plugins.PluginException import PluginException
from noora.system.ClassLoader import ClassLoader
from noora.io.File import File
from noora.plugins.postgresql.PGSQLPlugin import PGSQLPlugin
from noora.system import PropertyHelper
from noora.plugins.Fail import Fail


def plan(installed_version: str, prepared_arguments: dict) -> dict:
    """
    Create a deployment plan.
    """
    create_version = prepared_arguments.get("default_version")
    update_versions = prepared_arguments.get("update_versions")

    # a fresh, first time deployment
    if not installed_version:
        return {
            "default_version": create_version,
            "update_versions": update_versions
        }

    # the create is installed but not any updates
    if installed_version == create_version:
        return {
            "default_version": None,
            "update_versions": update_versions
        }

    # an update is installed, find where to start from
    if installed_version in update_versions:
        return {
            "default_version": None,
            "update_versions": update_versions[update_versions.index(installed_version) + 1::]
        }


def fetch_installed_version(connector, executor, properties):
    plugin_dir = properties.get('plugin.dir')
    script = File(os.path.join(plugin_dir, 'postgresql', 'deploy', 'fetch_version.sql'))
    executor['script'] = script

    try:
        connector.execute(executor, properties)
        result = connector.get_result()
    except PluginException as e:
        result = ""
        if 'ERROR:  relation "application_properties" does not exist' not in e.__str__():
            raise e

    if result:
        result = result.replace('\r\n', '\n')
        result = result.replace('\r', '\n')
        result = result.split('\n')
        return result[1]


class DeployPlugin(PGSQLPlugin):
    """
    This class provides functionality to drop, create and update a database
    to a specified version.
    """
    def _validate_and_prepare(self, properties, arguments):
        """
        Prepare the version list here. All other preparation is handled by the
        plugins in the chain.
        """
        prepared_args = {
            'default_version': properties.get('default_version'),
            'update_versions': [],
        }

        host = arguments.get('host')
        Fail.fail_on_no_host(host)
        prepared_args['host'] = host

        prepared_args['connection_string'] = arguments.get('connection_string')

        target_version = arguments.get('version')

        # If the target version is the default, we stop here because create takes care of this
        if prepared_args['default_version'] == target_version:
            return prepared_args

        # Here we determine the initial version
        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()

        # retrieve all versions
        for version in versions.list():
            # First add this version if it's not the default one, because that's handled by create
            if version.get_value() != prepared_args['default_version']:
                prepared_args['update_versions'].append(version.get_value())

            # Then, if we reached the target version, we break
            if version.get_value() == target_version:
                break

        return prepared_args

    def execute(self, properties, arguments):
        """
        Deploy the database. Retrieves the installed versions from the postgreSQL.
        Will loop over all versions that are not installed.

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **version**: Desired target version;
            * **host**: Hostname to connect to;
            * **port**: Port to connect to (optional);
            * **environment**: Name of the environment (optional).
        """
        prepared_args = self._validate_and_prepare(properties, arguments)
        host = prepared_args['host']

        # retrieve the database connection credentials through
        # the commandline option --connection-string
        connection_string = prepared_args['connection_string']
        if connection_string:
            users = PropertyHelper.connection_credentials(connection_string)

        if not connection_string:
            # retrieve the user credentials for this database project.
            users = properties.get('postgresql_users')

            # try to retrieve the users from the credentials file, when no users are configured in
            # myproject.json.
            if not users:
                # retrieve the name of this database project, introduced in version 1.0.12
                profile = PropertyHelper.get_profile(properties)
                if profile:
                    users = profile.get('postgresql_users')

        connector = self.get_connector()
        version_database = properties.get('version_database')
        executor = PropertyHelper.get_postgres_properties(users, host, version_database)
        installed_version = fetch_installed_version(connector, executor, properties)

        planned_versions = plan(installed_version, prepared_args)
        if planned_versions.get("default_version"):
            # find and execute the create plugin.
            plugin = ClassLoader.load_class(ClassLoader.find_plugin(properties['technology'], 'create'))
            plugin.execute(properties, arguments)

        # find and execute the update plugin.
        plugin = ClassLoader.load_class(ClassLoader.find_plugin(properties['technology'], 'update'))
        # create the arguments and execute the plugin
        for version in planned_versions.get("update_versions"):
            arguments['version'] = version
            plugin.execute(properties, arguments)

        # show that nothing happened, when
        if not planned_versions.get("default_version") and not planned_versions.get("update_versions"):
            print("Nothing to deploy")



import click

from noora.plugins.postgresql.deploy.deploy_plugin import DeployPlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('-v', '--version', required=False, help='The version to recreate')
@click.option('--connection-string', required=False)
@click.pass_obj
def cli(props, host, environment, version, connection_string):
    """
    Deploy a postgreSQL database.
    """
    plugin = DeployPlugin()
    plugin.execute(props, {'host': host, 'environment': environment, 'version': version,
                           'connection_string': connection_string})



import os

from noora.system import Ora
from noora.system import PropertyHelper
from noora.io.File import File

from noora.plugins.mssql.MssqlPlugin import MssqlPlugin
from noora.plugins.Fail import Fail

from noora.connectors.ConnectionExecutor import ConnectionExecutor


class DropPlugin(MssqlPlugin):
    """Class providing functionality to drop (clear out) a database."""
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        host = arguments.get('host')
        Fail.fail_on_no_host(host)
        Fail.fail_on_invalid_host(host, properties)
        Fail.fail_on_blocked_hosts(host, properties)
        prepared_args['host'] = host

        environment = arguments.get('environment')
        default_environment = properties.get('default_environment')
        environment = Ora.nvl(environment, default_environment)
        Fail.fail_on_invalid_environment(environment, properties)
        prepared_args['environment'] = environment

        return prepared_args

    def execute(self, properties, arguments):
        """
        Drop a database after checking if schema and environment are
        valid values. Also check that host is not on the block list.

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **host**: The hostname to drop on;
            * **environment**: Environment to drop the database from.
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        host = prepared_args['host']
        schemes = properties.get('schemes')
        environment = prepared_args['environment']
        database = properties.get('database')
        objects = properties.get('drop_objects')

        # retrieve the user credentials for this database project.
        users = properties.get('mssql_users')

        # try to retrieve the users from the credentials file, when no users are configured in
        # myproject.json.
        if not users:
            # retrieve the name of this database project, introduced in version 1.0.12
            profile = PropertyHelper.get_profile(properties)
            if profile:
                users = profile.get('mssql_users')

        for schema in schemes:
            print("dropping schema '{schema}' in database '{db}' "
                  "on host '{host}' using environment '{env}'".format(
                    schema=schema, db=database, host=host, env=environment))

            executor = PropertyHelper.get_mssql_properties(users, host, database, schema)
            executor['database'] = database

            connector = self.get_connector()

            for obj in objects:
                folder = File(os.path.join(properties.get('plugin.dir'), 'mssql', 'drop', obj))
                ConnectionExecutor.execute(connector, executor, properties, folder)

            print("schema '{}' dropped.".format(schema))



class PreProcessor(object):
    """Template parser"""
    @staticmethod
    def parse(file, properties):
        """
        Parse the template and replace all keys if applicable.

        :type file: noora.io.File.File
        :param file: The file to process;
        :type properties: dict
        :param properties: All keys to replace in the file with their respective values;

        :return: The parsed file as a string.
        """
        f = open(file.get_url())
        stream = f.read()
        f.close()

        for key in properties.keys():
            if properties.get(key):
                stream = stream.replace("{" + key + "}", properties.get(key))

        return stream



import os


class File(object):
    """
    File management functionality. Initialize a file object with its path (url).
    """
    def __init__(self, url=None):
        """Initialize the file path"""
        self.__url = url

    def get_url(self):
        """Return the path to the file"""
        return self.__url

    def is_file(self):
        """
        Returns True if url points to a file, false if not (e.g. when the
        file is a directory).
        """
        url = self.get_url()
        return os.path.isfile(url)

    def is_dir(self):
        """
        Returns True if url points to a directory, False otherwise.
        """
        url = self.get_url()
        return os.path.isdir(url)

    def tail(self):
        """
        Returns the tail part after splitting the File's url.
        """
        url = self.get_url()
        head, tail = os.path.split(url)
        return tail

    def head(self):
        """
        Returns the head part after splitting the File's url.
        """
        url = self.get_url()
        head, tail = os.path.split(url)
        return head

    def extension(self):
        """
        Get the file's extension
        """
        url = self.get_url()
        filename, extension = os.path.splitext(url)
        return extension

    def exists(self):
        """
        Check if the file exists.
        """
        if self.is_file():
            return True
        if self.is_dir():
            return True
        return False



import os
import subprocess


class StartupInfoFactory(object):
    """Factory for creating startup info instances"""
    @staticmethod
    def new_startup_info():
        startup_info = None
        if os.name == 'nt':
            startup_info = subprocess.STARTUPINFO()
            # USESHOWWINDOW
            startup_info.dwFlags |= 1
            # SW_HIDE
            startup_info.wShowWindow = 0
        return startup_info



import click

from .BuildPlugin import BuildPlugin


@click.command()
@click.option('-v', '--version', required=True)
@click.option('-d', '--database', required=False)
@click.pass_obj
def cli(props, version, database):
    """
    Build a package for this MySQL database.
    """
    plugin = BuildPlugin()
    plugin.execute(props, {'version': version, 'database': database})



import click

from noora.plugins.mssql.create.CreatePlugin import CreatePlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost', help='The host to deploy to')
@click.option('-e', '--environment', required=False, help='The environment to deploy in')
@click.pass_obj
def cli(props, host, environment):
    """
    Initialize an MSSQL database at the default version.
    """
    plugin = CreatePlugin()
    plugin.execute(props, {'host': host, 'environment': environment})



import click

from noora.plugins.mysql.drop.DropPlugin import DropPlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-d', '--database', required=False, help='Database name')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('-a', '--alias', required=False, help='Alias. Overrules the database option')
@click.pass_obj
def cli(props, host, database, environment, alias):
    """
    Drop (clear out) a MySQL database.
    """
    plugin = DropPlugin()
    plugin.execute(
        props,
        {'host': host, 'database': database, 'environment': environment, 'alias': alias},
    )



import os
from os.path import expanduser
import json

import noora
from noora.io.File import File


class Properties(object):
    """
    The properties object analyzes the environment and looks for project
    settings. The properties instance emulates the dict type.
    """
    def __init__(self):
        noora_dir = os.path.dirname(noora.__file__)
        current_dir = os.path.abspath('.')

        self.__props = {
            "noora.dir": noora_dir,
            "current.dir": current_dir,
            "plugin.dir": os.path.join(noora_dir, 'plugins'),
            "project.file": "myproject.json",
            'home.dir': expanduser('~'),
        }

        self.update_config()

    def __get_config(self):
        """
        Load a myproject.json file and update some extra parameters if this is an actual project
        """
        current_dir = self.__props.get("current.dir")
        project_file = self.__props.get("project.file")

        config_file = File(os.path.join(current_dir, project_file))
        if not config_file.exists():
            noora_dir = self.__props.get("noora.dir")
            config_file = File(os.path.join(noora_dir, project_file))

        # Read project configuration
        with open(config_file.get_url()) as fd:
            data = json.load(fd)

        return data

    def update_config(self):
        """
        Update the project configuration
        """
        # Get the configuration and merge it into the properties
        self.__props.update(self.__get_config())

        # If this is an actual project, update some other properties
        if 'project' in self:
            self['alter.dir'] = os.path.join(self['current.dir'], 'alter')
            self['create.dir'] = os.path.join(self['current.dir'], 'create')

            # Guess the database technology if not present
            if 'technology' not in self:
                self['technology'] = self['plugins'][0].split(".")[2]

    def __getitem__(self, item):
        # Do not catch exceptions here, but upstream
        return self.__props[item]

    def get(self, item, default=None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def __setitem__(self, key, value):
        self.__props[key] = value

    def __delitem__(self, key):
        del self.__props[key]

    def __contains__(self, item):
        return item in self.__props

    def keys(self):
        return self.__props.keys()



import os
from zipfile import ZipFile
from zipfile import ZIP_DEFLATED

from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader
from noora.version.Version import Version

from noora.system import Ora
from noora.io.File import File
from noora.io.Files import Files

from noora.plugins.mysql.MysqlPlugin import MysqlPlugin
from noora.plugins.Fail import Fail


class BuildPlugin(MysqlPlugin):
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        version = arguments.get('version')
        Fail.fail_on_no_version(version)
        Fail.fail_on_unknown_version(version, properties)
        prepared_args['version'] = version

        database = arguments.get('database')
        default_databases = properties.get('databases')
        databases = Ora.nvl(database, default_databases)
        Fail.fail_on_invalid_database(database, properties)
        prepared_args['databases'] = databases

        return prepared_args

    def execute(self, properties, arguments):
        """
        Build the package after verifying version is valid, and database,
        if provided.

        :param properties: The project properties
        :param arguments: A dict of {
            'version': 'The version to package for'
            'database': 'The database to package for (optional)'
        }
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        version = prepared_args['version']
        databases = prepared_args['databases']
        current_dir = properties.get('current.dir')
        component_name = properties.get('component_name')
        target_dir = os.path.join(current_dir, properties.get('component_target_folder'))

        objects = properties.get('create_objects')

        if version == properties.get("default_version"):
            build_dir = properties.get("create.dir")
        else:
            build_dir = os.path.join(properties.get("alter.dir"), version)

        # exclude the file 'version.sql', this file is excluded from the dat listing below.
        component_excluded_files = properties.get('component_excluded_files')
        excluded_files = properties.get('excluded_files')
        excluded_files.extend(component_excluded_files)

        # create the target folder, if not present.
        if not File(target_dir).exists():
            os.makedirs(target_dir)

        print("building component with version '{}'".format(version))

        zip_file = os.path.join(target_dir, component_name + '_' + version + '.zip')
        zip_handle = ZipFile(zip_file, 'w')

        for database in databases:
            for obj in objects:
                if not obj == 'lib':
                    # global ddl objects
                    folder = File(os.path.join(build_dir, database, 'ddl', obj))
                    zip_dir = File(os.path.join(component_name + '_' + version, 'ddl', obj))

                    files = Files.list_filtered(folder, properties)
                    for file in files:
                        print(file.get_url())
                        target_file = File(os.path.join(zip_dir.get_url(), file.tail()))
                        zip_handle.write(file.get_url(), target_file.get_url(), ZIP_DEFLATED)

            # global dat files
            folder = File(os.path.join(build_dir, database, 'dat'))
            zip_dir = File(os.path.join(component_name + '_' + version, 'dat'))

            files = Files.list_filtered(folder, properties)
            for file in files:
                print(file.get_url())
                target_file = File(os.path.join(zip_dir.get_url(), file.tail()))
                zip_handle.write(file.get_url(), target_file.get_url(), ZIP_DEFLATED)

        # create the version script in the dat folder
        if version == properties.get("default_version"):
            version_statement = properties.get("component_insert_statement")
        else:
            version_statement = properties.get("component_update_statement")
        version_statement = version_statement.replace('<version>', version)
        version_statement = version_statement.replace('<name>', component_name)

        f = open('version.sql', 'w')
        f.write(version_statement)
        f.close()

        zip_dir = File(os.path.join(component_name + '_' + version, 'dat'))
        target_file = File(os.path.join(zip_dir.get_url(), 'version.sql'))
        zip_handle.write('version.sql', target_file.get_url())

        # remove the version.sql file.
        os.remove('version.sql')

        # create the checkversion script in the root folder
        # first retrieve the previous version for the checkversion script
        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()
        previous = versions.previous(Version(version))

        component_select_statement = properties.get("component_select_statement")
        component_select_statement = component_select_statement.replace('<name>', component_name)
        component_select_statement = component_select_statement.replace(
            '<previous>', previous.get_value())

        print(component_select_statement)

        f = open('checkversion.sql', 'w')
        f.write(component_select_statement)
        f.close()

        zip_dir = File(os.path.join(component_name + '_' + version))
        target_file = File(os.path.join(zip_dir.get_url(), 'checkversion.sql'))
        zip_handle.write('checkversion.sql', target_file.get_url())

        # remove the version.sql file.
        os.remove('checkversion.sql')

        zip_handle.close()

        print("component with version {} created.".format(version))



from noora.io.Filters import Filters
from noora.io.File import File
from noora.io.Files import Files
from noora.io.FileFolderFilter import FileFolderFilter
from noora.io.FileExtensionFilter import FileExtensionFilter
from noora.io.FileFilter import FileFilter


class ConnectionExecutor(object):
    """
    This class unlocks functionality to execute a list of files in a
    specified folder, supporting file, folder and extension filtering.
    """
    @staticmethod
    def execute(connector, executor, properties, folder):
        """
        Execute all applicable files in the specified folder.

        :type connector: noora.connectors.Connector.Connector
        :param connector: A connector instance;
        :type executor: dict
        :param executor: Settings and credentials for the connector;
        :type properties: noora.system.Properties.Properties
        :param properties: Project properties;
        :param folder: The target folder to execute.
        """
        filters = Filters()

        excluded_folders = properties.get('excluded_folders')
        for excluded_folder in excluded_folders:
            ef = File(excluded_folder)
            ff = FileFolderFilter(ef)
            filters.add(ff)

        excluded_extensions = properties.get('excluded_extensions')
        for excluded_extension in excluded_extensions:
            ef = File("*." + excluded_extension)
            ff = FileExtensionFilter(ef)
            filters.add(ff)

        excluded_files = properties.get('excluded_files')
        for excluded_file in excluded_files:
            ef = File(excluded_file)
            ff = FileFilter(ef)
            filters.add(ff)

        if folder.is_dir():
            files = Files.list(folder)
            for file in files:
                if not filters.accept(file) and not file.is_dir():
                    url = file.get_url()
                    print(url)
                    executor['script'] = file
                    connector.execute(executor, properties)



import click

from noora.plugins.mysql.recreate.RecreatePlugin import RecreatePlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-d', '--database', required=False, help='Database name')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('-a', '--alias', required=False, help='Alias. Overrules the database option')
@click.option('-v', '--version', required=False, help='The version to recreate')
@click.pass_obj
def cli(props, host, database, environment, alias, version):
    """
    Drop and recreate a MySQL database to the specified or latest version.
    """
    plugin = RecreatePlugin()
    plugin.execute(
        props,
        {
            'host': host, 'database': database, 'environment': environment,
            'alias': alias, 'version': version
        },
    )



from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader

from noora.system.ClassLoader import ClassLoader

from noora.plugins.mssql.MssqlPlugin import MssqlPlugin


class RecreatePlugin(MssqlPlugin):
    """
    This class provides functionality to drop, create and update a database
    to a specified version.
    """
    def _validate_and_prepare(self, properties, arguments):
        """
        Prepare the version list here. All other preparation is handled by the
        plugins in the chain.
        """
        prepared_args = {
            'default_version': properties.get('default_version'),
            'update_versions': [],
        }
        target_version = arguments.get('version')

        # If the target version is the default, we stop here because create takes care of this
        if prepared_args['default_version'] == target_version:
            return prepared_args

        # Here we determine the initial version
        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()

        # retrieve all versions
        for version in versions.list():
            # First add this version if it's not the default one, because that's handled by create
            if version.get_value() != prepared_args['default_version']:
                prepared_args['update_versions'].append(version.get_value())

            # Then if we reached the target version, we break
            if version.get_value() == target_version:
                break

        return prepared_args

    def execute(self, properties, arguments):
        """
        Drop, create and update the database. Will loop over all versions and
        initialize the database to the latest version available.

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **version**: Desired target version;
            * **host**: Hostname to connect to;
            * **port**: Port to connect to (optional);
            * **environment**: Name of the environment (optional).
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        # find and execute the drop plugin.
        plugin = ClassLoader.load_class(ClassLoader.find_plugin(properties['technology'], 'drop'))
        plugin.execute(properties, arguments)

        # find and execute the create plugin.
        plugin = ClassLoader.load_class(ClassLoader.find_plugin(properties['technology'], 'create'))
        plugin.execute(properties, arguments)

        # find and execute the create plugin.
        plugin = ClassLoader.load_class(ClassLoader.find_plugin(properties['technology'], 'update'))

        # create the arguments and execute the plugin
        for version in prepared_args['update_versions']:
            arguments['version'] = version
            plugin.execute(properties, arguments)



import click

from noora.plugins.mssql.generate.GeneratePlugin import GeneratePlugin


@click.command()
@click.option('-v', '--version', required=False)
@click.pass_obj
def cli(props, version):
    """
    Bootstrap a new version of an MSSQL database project
    """
    plugin = GeneratePlugin()
    plugin.execute(props, {'version': version})


@click.command()
@click.option('-h', '--host', required=False, prompt=True, default='localhost')
@click.option('-p', '--port', required=False, prompt=True, type=int, default=1433)
@click.option('-d', '--database', required=True, prompt='Database name')
@click.option('-s', '--schema', required=True, prompt='Schema name')
@click.option('-U', '--username', required=True, prompt='Database username')
@click.option('-P', '--password', required=True, prompt='Database password',
              hide_input=True, confirmation_prompt=True)
@click.option('-v', '--version', required=False,
              prompt='Initial project version', default='1.0.0')
@click.pass_obj
def cli_outside_scope(props, host, port, database, schema, username, password, version):
    """
    Generate a new MSSQL database project
    """
    plugin = GeneratePlugin()
    plugin.execute(
        props,
        {
            'host': host, 'port': port, 'database': database, 'schema': schema,
            'username': username, 'password': password, 'version': version,
        },
    )



import os

from noora.version.Version import Version


class VersionGuesser(object):
    """Utilities for inspecting current version of a project"""
    def __init__(self, properties, versions):
        """
        Initialize the class.

        :param properties: An initialized properties object;
        :param versions: The list of versions for the project.
        """
        self.__properties = properties
        self.__versions = versions

    def guess(self, version=None):
        """
        Do some guessing based on the current version.

        :param version (optional): if provided, find the version in the list.
            If not provided, find the next version on the project;
        :return: The current (if found) or next version, or the default
            version if all else fails.
        """
        versions = self.__versions
        properties = self.__properties
        if version:
            return Version(version)
        if versions.next():
            return versions.next()
        return Version(properties.get("default_version"))

    def to_folder(self, version):
        properties = self.__properties
        if version == properties.get("default_version"):
            return properties.get("create.dir")
        return os.path.join(properties.get("alter.dir"), version)



import os

from noora.system import PropertyHelper
from noora.system import Ora
from noora.io.File import File

from noora.plugins.postgresql.PGSQLPlugin import PGSQLPlugin
from noora.plugins.Fail import Fail

from noora.connectors.ConnectionExecutor import ConnectionExecutor


class CreatePlugin(PGSQLPlugin):
    """Plugin for creating (initializing) a PostgreSQL database."""
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        host = arguments.get('host')
        Fail.fail_on_no_host(host)
        Fail.fail_on_invalid_host(host, properties)
        prepared_args['host'] = host

        environment = arguments.get('environment')
        default_environment = properties.get('default_environment')
        environment = Ora.nvl(environment, default_environment)
        Fail.fail_on_invalid_environment(environment, properties)
        prepared_args['environment'] = environment

        prepared_args['connection_string'] = arguments.get('connection_string')

        return prepared_args

    def execute(self, properties, arguments):
        """
        Create a new database instance for the initial version.

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **host**: The hostname where the database is running;
            * **environment**: The environment to create the database in (optional).
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        host = prepared_args['host']
        environment = prepared_args['environment']
        databases = properties.get('databases')

        objects = properties.get('create_objects')

        # retrieve the database connection credentials through
        # the commandline option --connection-string
        connection_string = prepared_args['connection_string']
        if connection_string:
            users = PropertyHelper.connection_credentials(connection_string)

        if not connection_string:
            # retrieve the user credentials for this database project.
            users = properties.get('postgresql_users')

            # try to retrieve the users from the credentials file, when no users are configured in
            # myproject.json.
            if not users:
                # retrieve the name of this database project, introduced in version 1.0.12
                profile = PropertyHelper.get_profile(properties)
                if profile:
                    users = profile.get('postgresql_users')

        # fail when no users are found. This means that they are not set in myproject.json or
        # credentials.json
        Fail.fail_on_no_users(users)

        connector = self.get_connector()
        create_dir = properties.get('create.dir')

        for database in databases:
            print("creating database '{db}' on host '{host}' using environment '{env}'".format(
                db=database, host=host, env=environment))

            executor = PropertyHelper.get_postgres_properties(users, host, database)

            for obj in objects:
                # global ddl objects
                folder = File(os.path.join(create_dir, database, 'ddl', obj))
                ConnectionExecutor.execute(connector, executor, properties, folder)

                # environment specific ddl objects
                folder = File(os.path.join(create_dir, database, 'ddl', obj, environment))
                ConnectionExecutor.execute(connector, executor, properties, folder)

            # global dat objects
            folder = File(os.path.join(create_dir, database, 'dat'))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            # environment specific dat objects
            folder = File(os.path.join(create_dir, database, 'dat', environment))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            print("database '{}' created.".format(database))



import click

from noora.plugins.mssql.recreate.RecreatePlugin import RecreatePlugin


@click.command()
@click.option(
    '-h', '--host', required=True, default='localhost', help='The host to recreate the database on')
@click.option(
    '-e', '--environment', required=False, help='The environment to recreeate the database in')
@click.option('-v', '--version', required=False, help='The target version to recreate')
@click.pass_obj
def cli(props, host, environment, version):
    """
    Drop and recreate an MSSQL database to the specified or latest version.
    """
    plugin = RecreatePlugin()
    plugin.execute(
        props, {'host': host, 'environment': environment, 'version': version})



def nvl(*args):
    """
    Return the first argument, or the second argument if the first was not provided. Serves
    as default value selector.
    """
    count = len(args)
    if count != 2:
        raise ValueError("invalid number of arguments")

    if args[0]:
        return args[0]
    return args[1]


# TODO: this method is not used in the project. Maybe remove?
def nvl2(*args):
    count = len(args)
    if count != 3:
        raise ValueError("invalid number of arguments")

    if args[0]:
        return args[1]
    return args[2]


# TODO: this method is not used in the project. Maybe remove?
def nnvl(*args):
    count = len(args)
    if count != 2:
        raise ValueError("invalid number of arguments")

    if not args[0]:
        return args[0]
    return args[1]


# TODO: this method is not used in the project. Maybe remove?
def nnvl2(*args):
    count = len(args)
    if count != 3:
        raise ValueError("invalid number of arguments")

    if not args[0]:
        return args[1]
    return args[2]


# TODO: this method is not used in the project. Maybe remove?
def decode(*args):
    count = len(args)
    if count < 3:
        raise ValueError("not enough arguments for method")
    key = args[0]
    for i in range(1, count - 1, 2):
        if key == args[i]:
            return args[i+1]
    if count % 2 == 0:
        return args[count-1]


# TODO: this method is not used in the project. Maybe remove?
def nnvl_decode(*args):
    count = len(args)
    if count < 3:
        raise ValueError("not enough arguments for method")
    for i in range(0, count - 1, 2):
        if args[i]:
            return args[i+1]
    if not count % 2 == 0:
        return args[count-1]


# TODO: this method is not used in the project. Maybe remove?
def iif(*args):
    count = len(args)
    if count < 3:
        raise ValueError("not enough arguments for method")
    if args[1] in args[0]:
        return args[0][args[1]]
    return args[2]



from noora.io.File import File
from noora.io.Files import Files
from noora.version.Version import Version


class VersionLoader(object):
    """
    Class providing functionality to determine all versions on the current
    project.
    """
    def __init__(self, versions):
        """
        Initialize the class.

        :param versions: A ``noora.version.Versions.Versions`` instance.
        """
        self.__versions = versions

    def load(self, properties):
        """
        Find all versions on the current project and add them to the list.

        :param properties: An instance of ``noora.system.Properties.Properties``.
        """
        alter = File(properties.get("alter.dir"))
        if alter.exists():
            files = Files()
            for version in files.list(alter):
                self.__versions.add(Version(version.tail()))

        create = File(properties.get("create.dir"))
        if create.exists():
            self.__versions.add(Version(properties.get("default_version")))



class Plugin(object):
    """
    This is the base class to be used by all plugins.
    """
    _connectable = None

    def __init__(self, connectable=None):
        """
        Initialize the class.

        :param connectable: If set, initializes the connectable and sets this
            as connector on the class instead of the default connectable class.
            This allows you to override a connector for a technology without
            having to define a new plugin class.
        """
        if connectable:
            self._connector = connectable()
        else:
            self._connector = self._connectable()

    def get_connector(self):
        return self._connector

    def _validate_and_prepare(self, properties, arguments):
        """
        Internal method called by execute to validate and prepare the arguments.

        :param properties: An initialized noora.system.Properties.Properties class
        :param arguments: A dict containing the arguments for execution
        """
        raise NotImplementedError("Plugin._validate_and_prepare should be defined by subclasses")

    def execute(self, properties, arguments):
        """
        The execute method. This should be implemented by subclasses.

        :param properties: An initialized noora.system.Properties.Properties class
        :param arguments: A dict containing the arguments for execution
        """
        raise NotImplementedError("Plugin.execute should be defined by subclasses")



import click

from noora.plugins.postgresql.update.UpdatePlugin import UpdatePlugin


@click.command()
@click.option('-v', '--version', required=False)
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('--connection-string', required=False)
@click.pass_obj
def cli(props, version, host, environment, connection_string):
    """
    Update a MSSQL database to the specified version.
    """
    plugin = UpdatePlugin()
    plugin.execute(props, {'version': version, 'host': host, 'environment': environment,
                           'connection_string': connection_string})



class Filters(object):
    def __init__(self):
        self.__filters = []

    def add(self, filterable):
        self.__filters.append(filterable)

    def list(self):
        return self.__filters

    def accept(self, fileable):
        filters = self.__filters
        for filter in filters:
            if filter.accept(fileable):
                return True
        return False



import os

from noora.io.File import File
from noora.io.Filters import Filters
from noora.io.FileFolderFilter import FileFolderFilter
from noora.io.FileExtensionFilter import FileExtensionFilter
from noora.io.FileFilter import FileFilter


class Files(object):
    # FIXME: docstrings
    @staticmethod
    def list_filtered(folder, properties):
        filters = Filters()
        results = []

        excluded_folders = properties.get('excluded_folders')
        for excluded_folder in excluded_folders:
            ef = File(excluded_folder)
            ff = FileFolderFilter(ef)
            filters.add(ff)

        excluded_extensions = properties.get('excluded_extensions')
        for excluded_extension in excluded_extensions:
            ef = File("*." + excluded_extension)
            ff = FileExtensionFilter(ef)
            filters.add(ff)

        excluded_files = properties.get('excluded_files')
        for excluded_file in excluded_files:
            ef = File(excluded_file)
            ff = FileFilter(ef)
            filters.add(ff)

        if folder.is_dir():
            files = Files.list(folder)
            for file in files:
                if not filters.accept(file) and not file.is_dir():
                    results.append(file)
        return results

    @staticmethod
    def list(file=None, recursive=False, exclude=None):
        result = []

        if file.exists():
            folder = file.get_url()
            file_list = os.listdir(folder)
            file_list.sort()
            for file_item in file_list:
                url = os.path.join(folder, file_item)
                candidate_file = File(url)
                if candidate_file.is_file() and exclude != "file":
                    result.append(candidate_file)
                if candidate_file.is_dir() and exclude != "directory":
                    result.append(candidate_file)
                if candidate_file.is_dir() and recursive == True:
                    recursive_files = Files.list(candidate_file, recursive, exclude)
                    for recursive_file in recursive_files:
                        result.append(recursive_file)
        return result



import os
from importlib import import_module

import click


class GenerateCommand(click.MultiCommand):
    def list_commands(self, ctx):
        """
        Return a list of technologies that contain a 'generate' package with
        a cli_outside_scope method

        :return: A list of available technologies
        """
        res = []
        plugin_dir = ctx.obj['plugin.dir']
        for f in os.listdir(plugin_dir):
            f_path = "{}/{}".format(plugin_dir, f)
            if os.path.isdir(f_path) and os.path.exists("{}/generate".format(f_path)):
                try:
                    mod = import_module("noora.plugins.{}.generate".format(f))
                    if hasattr(mod, 'cli_outside_scope'):
                        res.append(f)
                # Generic handling, Py < 3.6 throws ImportError, >= 3.6 ModuleNotFoundError
                except:
                    pass
        return res

    def get_command(self, ctx, cmd_name):
        try:
            mod = import_module("noora.plugins.{}.generate".format(cmd_name))
        # Generic handling, Py < 3.6 throws ImportError, >= 3.6 ModuleNotFoundError
        except:
            return None

        if hasattr(mod, 'cli_outside_scope'):
            return mod.cli_outside_scope

        return None


@click.command(cls=GenerateCommand)
def cli():
    """
    The generate plugin can be used to create a new database project or
    bootstrap a new version for the currently selected project.
    """
    pass



from noora.shell.StartupInfoFactory import StartupInfoFactory


class CallFactory(object):
    """Factory for creating call instances"""
    @staticmethod
    def new_call(args):
        call = {
            'args': args,
            'shell': True,
            'stdin': None,
            'stderr': None,
            'startupinfo': StartupInfoFactory.new_startup_info(),
        }
        return call



import os
from importlib import import_module

import click

from noora.system.Properties import Properties


class App(click.MultiCommand):
    @staticmethod
    def initialize_and_get_properties(ctx):
        """
        Makes sure the Properties object is initialized and assigned to the context object.
        :param ctx: The click context
        :return: a reference to the properties object
        """
        if not ctx.obj:
            ctx.obj = Properties()
        return ctx.obj

    def list_commands(self, ctx):
        """
        Get the commands based on the context. If 'project' is defined on the
        properties, look up all available plugins for the project's
        technology. Else, only return 'generate'.

        :return: A list of available commands
        """
        props = self.initialize_and_get_properties(ctx)

        if 'project' in props:
            technology = props.get('technology')
            res = []

            # Loop over the plugins for this technology and add all plugins to the list that have
            # a valid execute method # configured.
            plugin_path = os.path.join(props['plugin.dir'], technology)
            for p in os.listdir(plugin_path):
                p_path = os.path.join(plugin_path, p)
                if os.path.isdir(p_path):
                    # Try to import the cli command from the package root
                    try:
                        mod = import_module('noora.plugins.{}.{}'.format(technology, p))
                        if hasattr(mod, 'cli'):
                            res.append(p)
                    except:
                        pass
            return res

        # NOTE: in the future it might happen we support more than just 'generate' outside of a
        # project's scope. In this case one should loop over all technologies and plugins per
        # technology to check if an execute method exists that can be run outside of a project's
        # scope.
        else:
            return ['generate', ]

    def get_command(self, ctx, cmd_name):
        props = self.initialize_and_get_properties(ctx)

        if 'project' in props:
            # Import the module and return the cli method
            try:
                mod = import_module('noora.plugins.{}.{}'.format(props['technology'], cmd_name))
            except:
                return None
            if hasattr(mod, 'cli'):
                return mod.cli
        elif cmd_name == 'generate':
            mod = import_module('noora.system.GenerateCommand')
            return mod.cli

        return None



import subprocess

from noora.exceptions.plugins.PluginException import PluginException


class Shell(object):
    """Shell interaction"""
    @staticmethod
    def execute(call):
        """
        Execute the command in ``call``. Returns the standard output if
        execution was successful, raises an exception with results from
        standard error otherwise.

        :param call: A dictionary, containing:

            * **args**: The command to execute as a string;
            * **stdout**: File-like object to write standard output to (optional);
            * **stderr**: File-like object to write standard error to (optional);
            * **stdin**: File-like object to read standard input from (optional);
            * **startupinfo**: Extra flags (optional).

        :return: The result from standard output.
        """
        args = call['args']
        shell = call['shell']
        stdout = call['stdout']
        stderr = call['stderr']
        stdin = call['stdin']
        startupinfo = call['startupinfo']

        # TODO: Popen isn't used any more. Remove?
        #p = subprocess.Popen(args, shell=shell, stdin=stdin, stderr=stderr, startupinfo=startupinfo)

        result = subprocess.call(
            args, shell=True, stdin=stdin, stdout=stdout, stderr=stderr, startupinfo=startupinfo)

        #output = p.communicate()
        #if p.returncode != 0:

        if result != 0:
            f = open(stderr.name)
            stream = f.read()
            f.close()
            raise PluginException(stream)

        f = open(stdout.name)
        stream = f.read()
        f.close()

        return stream



from importlib import import_module


class ClassLoader(object):
    @staticmethod
    def find_plugin(technology, name):
        """
        Find a plugin based on the technology and the name. Note that this is
        a 'dumb' method that only builds up the module path.

        :param technology: Technology to look in for the plugin
        :param name: Name of the plugin
        :return: A module path including class name
        """
        plugin_class = '{}Plugin'.format(name.lower().capitalize())

        return 'noora.plugins.{}.{}.{}.{}'.format(technology, name, plugin_class, plugin_class)

    @staticmethod
    def load_class(pattern):
        """
        Import the class from the pattern and return an instance of the class.
        :param pattern: Full python module path including the class name

        :return: Class instance
        """
        pattern_list = pattern.rsplit(".", 1)
        if len(pattern_list) < 2:
            raise ValueError(
                'A class pattern should consist of at least one module and a classname')

        mod_path = pattern_list[0]
        class_name = pattern_list[1]

        mod = import_module(mod_path)
        cls = getattr(mod, class_name)
        instance = cls()

        return instance



from noora.io.File import File
from noora.shell.Shell import Shell
from noora.shell.CallFactory import CallFactory

from noora.processor.PreProcessor import PreProcessor
from noora.connectors.Connector import Connector


class MssqlConnector(Connector):
    """
    Connector for Microsoft SQL server.
    """
    def execute(self, executable, properties):
        """
        Execute the script provided by `executable` on the target server.

        :type executable: dict
        :param executable: Should contain the following keys and values:

            * **host**: The address of the server to connect to;
            * **port**: Server port to connect to;
            * **database**: The database name;
            * **schema**: The schema to use;
            * **username**: Database username;
            * **password**: Database user password;
            * **script**: Path to the script to execute.
        :type properties: noora.system.Properties.Properties
        :param properties: A Noora project properties instance
        """
        script = executable['script']

        cp = dict()
        cp['database'] = executable['database']
        if 'schema' in executable.keys():
            cp['schema'] = executable['schema']
        if 'environment' in properties.keys():
            cp['environment'] = properties.get('environment')
        if 'previous' in properties.keys():
            cp['previous'] = properties.get('previous')

        stream = PreProcessor.parse(script, cp)

        tmp = File("tmp.sql")
        f = open(tmp.get_url(), 'w')
        f.write(stream)
        f.close()

        # FIXME: remove?
        #script_reader = open(tmp.get_url())

        feedback = File('feedback.log')
        feedback_writer = open(feedback.get_url(), 'w')
        # FIXME: rewrite with format
        statement = "sqlcmd -b -S {host},{port} -U {user} -P {passwd} -d {db} -i {url}".format(
            host=executable['host'],
            port=executable.get('port') or 1433,
            user=executable['username'],
            passwd=executable['password'],
            db=executable['database'],
            url=tmp.get_url(),
        )

        call = CallFactory.new_call(statement)
        call['stdout'] = feedback_writer
        call['stderr'] = feedback_writer
        result = Shell.execute(call)
        self.set_result(result)




import six

from noora.exceptions.plugins.PluginException import PluginException
from noora.exceptions.plugins.BlockedHostException import BlockedHostException
from noora.exceptions.plugins.UnknownVersionException import UnknownVersionException

from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader
from noora.version.Version import Version


class Fail(object):
    """
    Class providing assertion methods for various attributes.
    PluginExceptions will be raised when an assertion fails.
    """
    @staticmethod
    def fail_on_no_host(host):
        """Verify that ``host`` is a string"""
        if not host:
            raise PluginException("no host was given")
        elif not issubclass(type(host), six.string_types):
            raise PluginException("provided host is not a string")

    @staticmethod
    def fail_on_no_port(port):
        """Verify that ``port`` is an integer"""
        if not port:
            raise PluginException("no port was provided")
        elif type(port) is not int:
            raise PluginException("provided port is not an integer")

    @staticmethod
    def fail_on_no_version(version):
        """Verify that ``version`` is a string"""
        if not version:
            raise PluginException("no version was given")
        elif not issubclass(type(version), six.string_types):
            raise PluginException("version is not a string")

    @staticmethod
    def fail_on_no_users(users):
        """Verify that ``users`` is a list"""
        if not users:
            raise PluginException("no users found")
        elif type(users) is not list:
            raise PluginException("provided users is not a list")

    @staticmethod
    def fail_on_no_schema(schema):
        """Verify that ``schema`` is a string"""
        if not schema:
            raise PluginException("no schema name provided")
        elif not issubclass(type(schema), six.string_types):
            raise PluginException("provided schema is not a string")

    @staticmethod
    def fail_on_no_database(database):
        """Verify that ``database`` is a string"""
        if not database:
            raise PluginException("no database name provided")
        elif not issubclass(type(database), six.string_types):
            raise PluginException("provided database is not a string")

    @staticmethod
    def fail_on_no_user(user):
        """Verify that ``user`` is a string"""
        if not user:
            raise PluginException("no username provided")
        elif not issubclass(type(user), six.string_types):
            raise PluginException("provided username is not a string")

    @staticmethod
    def fail_on_invalid_database(database, properties):
        """Verify that ``database`` is valid for this project"""
        if database and database not in properties.get('databases'):
            raise PluginException("database '{}' is not valid for this project".format(database))

    @staticmethod
    def fail_on_invalid_schema(schema, properties):
        """Verify that ``schema`` is valid for this project"""
        if schema and schema not in properties.get('schemes'):
            raise PluginException("schema '{}' is not valid for this project".format(schema))

    @staticmethod
    def fail_on_invalid_environment(environment, properties):
        """Verify that ``environment`` is valid for this project"""
        if environment and environment not in properties.get('environments'):
            raise PluginException(
                "environment '{}' is not valid for this project".format(environment))

    @staticmethod
    def fail_on_invalid_alias(alias, properties):
        """Verify that ``alias`` is valid for this project"""
        if alias and alias not in properties.get('aliasses'):
            raise PluginException("alias '{}' is not valid for this project".format(alias))

    @staticmethod
    def fail_on_unknown_version(version, properties):
        """Verify that ``version`` is valid for this project"""
        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()

        if not versions.exists(Version(version)):
            raise UnknownVersionException(
                "version {} is not valid for this project".format(version))

    @staticmethod
    def fail_on_blocked_hosts(host, properties):
        """Make sure that ``host`` is not in the list of blocked hosts"""
        blocked_hosts = properties.get('blocked_hosts')
        if blocked_hosts:
            if host in blocked_hosts:
                raise BlockedHostException("blocked host: {}".format(host))

    @staticmethod
    def fail_on_invalid_host(host, properties):
        """
        Verify that host is in the host list for this project

        :param host: The host to look up
        :param host_list: Properties containing the host list
        """
        host_list = properties.get('{}_hosts'.format(properties.get('technology')))
        if host_list:
            if host not in host_list:
                raise PluginException(
                    "Host {} not in list of valid hosts for this project".format(host))



#!/usr/bin/env python
import unittest

from noora.io.File import File
from noora.processor.PreProcessor import PreProcessor


class TestPreProcessor(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testProcessor(self):
        script = File("drop_tables.sql")

        properties = dict()
        properties['database'] = 'orcl'

        stream = PreProcessor.parse(script, properties)
        self.assertEqual(stream, 'orcl', "invalid transformation")


if __name__ == '__main__':
    unittest.main()



import os
import shutil

from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader
from noora.version.VersionGuesser import VersionGuesser

from noora.plugins.mysql.MysqlPlugin import MysqlPlugin
from noora.plugins.Fail import Fail
from noora.io.File import File
from noora.io.Files import Files


class GeneratePlugin(MysqlPlugin):
    """
    This class provides generation functionality for MySQL projects.
    """
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        # If no hostname was provided, we only set the version and then return
        version = arguments.get('version')
        if version:
            prepared_args['version'] = version

        host = arguments.get('host')
        if not host:
            return prepared_args

        project_file = properties.get('project.file')
        current_dir = properties.get('current.dir')

        # Determine the project's name and create a new directory
        database = arguments.get('database')
        Fail.fail_on_no_database(database)
        project = database + "-db"

        # First check the rest of the arguments before we continue
        port = arguments.get('port')
        Fail.fail_on_no_port(port)
        username = arguments.get('username')
        Fail.fail_on_no_user('username')
        password = arguments.get('password')

        # Get the json project template for this plugin
        template_file = os.path.join(
            properties.get('plugin.dir'), 'mysql', 'generate', 'templates', project_file)

        # Create the project directory
        os.mkdir(project)

        # Parse the template and replace the parameters
        with open(template_file) as fd:
            stream = fd.read()
        stream = stream.replace('{project}', project)
        stream = stream.replace('{database}', database)
        stream = stream.replace('{host}', host)
        stream = stream.replace('{port}', str(port))
        stream = stream.replace('{username}', username)
        stream = stream.replace('{password}', password)
        stream = stream.replace('{version}', version)

        # Write the new configuration back to file
        config_file = os.path.join(current_dir, project, project_file)
        with open(config_file, 'w') as fd:
            fd.write(stream)

        # Update the project properties to set current.dir to the new project dir
        properties['current.dir'] = os.path.join(current_dir, project)

        # Update the project properties
        properties.update_config()

        return prepared_args

    def execute(self, properties, arguments):
        """
        Create or upgrade a project. The version is always required, the rest
        only when generating an entirely new project.

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **version**: (Initial) project version to generate for;
            * **host**: The hostname for the new project. If not provided, an upgrade is assumed (optional);
            * **port**: Port to connect to (optional);
            * **database**: Name of the database (optional);
            * **username**: Database username (optional);
            * **password**: Database password (optional).
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        template_dir = os.path.join(
            properties.get('plugin.dir'), 'mysql', 'generate', 'templates')

        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()
        version_guesser = VersionGuesser(properties, versions)
        next_version = version_guesser.guess(prepared_args.get('version')).to_string()
        version_dir = version_guesser.to_folder(next_version)

        # create the version folder
        os.makedirs(version_dir)

        databases = properties.get('databases')
        version_database = properties.get('version_database')
        default_version = properties.get('default_version')
        environments = properties.get('environments')
        objects = properties.get('create_objects')

        for database in databases:
            # create the scheme folder
            database_dir = os.path.join(version_dir, database)
            os.mkdir(database_dir)

            # create the dat folder
            dat_dir = os.path.join(database_dir, 'dat')
            os.mkdir(dat_dir)

            # create the version script in the dat folder
            if database == version_database:
                version_file = os.path.join(dat_dir, 'version.sql')
                f = open(version_file, 'w')

                if next_version == default_version:
                    stream = properties.get('version_insert_statement')
                else:
                    stream = properties.get('version_update_statement')

                stream = stream.replace('<version>', next_version)
                f.write(stream)
                f.close()

                # sqlScript=self.getSqlVersionStatement(versions, version)
                # projectHelper.writeFile(datFolder+os.sep+'version.sql', sqlScript)

            # create the environment folders in the dat folder
            for environment in environments:
                os.mkdir(os.path.join(dat_dir, environment))

                # create the environment script in the dat folder.
                if database == version_database and next_version == default_version:
                    environment_file = os.path.join(dat_dir, environment, 'environment.sql')

                    f = open(environment_file, 'w')
                    stream = properties.get('environment_insert_statement')
                    stream = stream.replace('<environment>', environment)
                    f.write(stream)
                    f.close()

            # create the ddl folder
            ddl_dir = os.path.join(database_dir, 'ddl')
            os.mkdir(ddl_dir)

            # create the object folders in the ddl folder
            for obj in objects:
                os.mkdir(os.path.join(ddl_dir, obj))

            # create the template code on create.
            if database == version_database and next_version == default_version:
                for obj in objects:
                    object_dir = os.path.join(template_dir, obj)
                    if os.path.exists(object_dir):
                        files = Files.list(File(object_dir))
                        for file in files:
                            shutil.copyfile(
                                file.get_url(), os.path.join(ddl_dir, obj, file.tail()))

        print("version {} created.".format(next_version))



from noora.io.File import File
from noora.shell.Shell import Shell
from noora.shell.CallFactory import CallFactory

from noora.connectors.Connector import Connector
from noora.processor.PreProcessor import PreProcessor


class MysqlConnector(Connector):
    """
    Connector for MySQL server.
    """
    def execute(self, executable, properties):
        """
        Execute the script provided by `executable` on the target server.

        :type executable: dict
        :param executable: Should contain the following keys and values:

            * **host**: The address of the server to connect to;
            * **port**: Server port to connect to;
            * **database**: The database name;
            * **username**: Database username;
            * **password**: Database user password;
            * **script**: Path to the script to execute.
        :type properties: noora.system.Properties.Properties
        :param properties: A Noora project properties instance
        """
        script = executable['script']

        cp = {
            'database': executable['database'],
        }

        if 'environment' in properties.keys():
            cp['environment'] = properties.get('environment')
        if 'previous' in properties.keys():
            cp['previous'] = properties.get('previous')

        stream = PreProcessor.parse(script, cp)

        tmp = File("tmp.sql")
        f = open(tmp.get_url(), 'w')
        f.write(stream)
        f.close()

        script_reader = open(tmp.get_url())

        feedback = File('feedback.log')
        feedback_writer = open(feedback.get_url(), 'w')

        statement = "mysql --show-warnings --port={port} --host={host} --user={user} " \
                    "--password={passwd} {db}".format(
                        host=executable['host'],
                        port=executable.get('port') or '3306',
                        user=executable['username'],
                        passwd=executable['password'],
                        db=executable['database'],
                    )
        call = CallFactory.new_call(statement)
        call['stdin'] = script_reader
        call['stdout'] = feedback_writer
        call['stderr'] = feedback_writer
        result = Shell.execute(call)
        self.set_result(result)




from io import StringIO

from noora.io.File import File
from noora.shell.Shell import Shell
from noora.shell.CallFactory import CallFactory

from noora.processor.PreProcessor import PreProcessor
from noora.connectors.Connector import Connector


class PGSQLConnector(Connector):
    """
    Connector for PostgreSQL server.
    """
    def execute(self, executable, properties):
        """
        Execute the script provided by `executable` on the target server.

        :type executable: dict
        :param executable: Should contain the following keys and values:

            * **host**: The address of the server to connect to;
            * **port**: Server port to connect to;
            * **database**: The database name;
            * **username**: Database username;
            * **password**: Database user password;
            * **script**: Path to the script to execute.
        :type properties: noora.system.Properties.Properties
        :param properties: A Noora project properties instance
        """
        script = executable['script']

        cp = {
            'database': executable['database'],
            'username': executable['username'].split('@')[0]
        }
        if 'environment' in properties.keys():
            cp['environment'] = properties.get('environment')
        if 'previous' in properties.keys():
            cp['previous'] = properties.get('previous')

        stream = PreProcessor.parse(script, cp)

        tmp = File("tmp.sql")
        f = open(tmp.get_url(), 'w')
        f.write(stream)
        f.close()

        script_reader = open(tmp.get_url())

        feedback = File('feedback.log')
        feedback_writer = open(feedback.get_url(), 'w')

        statement = \
            "PGPASSWORD={passwd} psql -h {host} -p {port} -U {user} -d {db} " \
            "-v ON_ERROR_STOP=1 --csv ".format(
                host=executable['host'],
                port=executable.get('port'),
                user=executable['username'],
                passwd=executable['password'],
                db=executable['database'],
            )

        call = CallFactory.new_call(statement)
        call['stdin'] = script_reader
        call['stdout'] = feedback_writer
        call['stderr'] = feedback_writer
        result = Shell.execute(call)
        self.set_result(result)




import os

from noora.system import PropertyHelper
from noora.system import Ora
from noora.io.File import File

from noora.plugins.mysql.MysqlPlugin import MysqlPlugin
from noora.plugins.Fail import Fail

from noora.connectors.ConnectionExecutor import ConnectionExecutor


class CreatePlugin(MysqlPlugin):
    """Plugin for creating (initializing) a MySQL database."""
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        host = arguments.get('host')
        Fail.fail_on_no_host(host)
        Fail.fail_on_invalid_host(host, properties)
        prepared_args['host'] = host

        environment = arguments.get('environment')
        default_environment = properties.get('default_environment')
        environment = Ora.nvl(environment, default_environment)
        Fail.fail_on_invalid_environment(environment, properties)
        prepared_args['environment'] = environment

        alias = arguments.get('alias')
        Fail.fail_on_invalid_alias(alias, properties)
        # if an alias is given, only this database will be installed, other databases will be
        # ignored.
        if alias:
            print("using alias: {}".format(alias))
            prepared_args['databases'] = [alias]
        else:
            database = arguments.get('database')
            Fail.fail_on_invalid_database(database, properties)
            default_databases = properties.get('databases')
            databases = Ora.nvl(database, default_databases)
            prepared_args['databases'] = databases

        return prepared_args

    def execute(self, properties, arguments):
        """
        Create a new database instance for the initial version

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **host**: The hostname where the database will run;
            * **database**: The database to create in (optional);
            * **environment** (optional): The environment to create the database in;
            * **alias** (optional): The database alias. If provided, this will overrule the database argument.
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        #: Some comment
        host = prepared_args['host']
        environment = prepared_args['environment']
        databases = prepared_args['databases']
        database_aliases = properties.get('database_aliases')

        objects = properties.get('create_objects')

        # retrieve the user credentials for this database project.
        users = properties.get('mysql_users')

        # try to retrieve the users from the credentials file, when no users are configured in
        # myproject.json.
        if not users:
            # retrieve the name of this database project, introduced in version 1.0.12
            profile = PropertyHelper.get_profile(properties)
            if profile:
                users = profile.get('mysql_users')

        # fail when no users are found. This means that they are not set in myproject.json or
        # credentials.json
        Fail.fail_on_no_users(users)

        connector = self.get_connector()
        create_dir = properties.get('create.dir')

        for database in databases:
            print("creating database '{db}' on host '{host}' using environment '{env}'".format(
                db=database, host=host, env=environment))

            executor = PropertyHelper.get_mysql_properties(users, host, database)

            database_folder = PropertyHelper.get_database_folder(database, database_aliases)

            for obj in objects:
                # global ddl objects
                folder = File(os.path.join(create_dir, database_folder, 'ddl', obj))
                ConnectionExecutor.execute(connector, executor, properties, folder)

                # environment specific ddl objects
                folder = File(os.path.join(create_dir, database_folder, 'ddl', obj, environment))
                ConnectionExecutor.execute(connector, executor, properties, folder)

            # global dat objects
            folder = File(os.path.join(create_dir, database_folder, 'dat'))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            # environment specific dat objects
            folder = File(os.path.join(create_dir, database_folder, 'dat', environment))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            print("database '{}' created.".format(database))



from noora.connectors.Connectable import Connectable


class Connector(Connectable):
    """
    Connector base class. All technology connectors should inherit this.
    """
    def __init__(self):
        self.__result = None

    def get_result(self):
        return self.__result

    def set_result(self, result):
        self.__result = result

    def execute(self, executable, properties):
        """Override this method in a subclass"""
        raise NotImplementedError("execute method not implemented")



import click

from noora.plugins.postgresql.drop.DropPlugin import DropPlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('--connection-string', required=False)
@click.pass_obj
def cli(props, host, environment, connection_string):
    """
    Drop (empty) a PostgreSQL database.
    """
    plugin = DropPlugin()
    plugin.execute(props, {'host': host, 'environment': environment, 'connection_string': connection_string})



class Version(object):
    """Inspection class for a version string."""
    def __init__(self, value):
        """
        Initialize the version by splitting its parts and calculating the weight.

        :param value: A version in the form of A[.B[.C[.D]]]
        """
        self.__value = value

        i = 0
        result = 0
        for item in value.split('.'):
            # FIXME: if you have more than 1000 versions of a version part,
            # this will give you bad times
            result = result + float(item) / 1000**i
            i = i + 1
        self.__weight = result

    def get_value(self):
        """Get version"""
        return self.__value

    def to_string(self):
        return self.__value

    def get_major(self):
        """Return major part of the version"""
        value = self.__value
        if len(value.split('.')) >= 1:
            return value.split('.')[0]

    def get_minor(self):
        """Return minor version part, if applicable"""
        value = self.__value
        if len(value.split('.')) >= 2:
            return value.split('.')[1]

    def get_revision(self):
        """Return revision part, if applicable"""
        value = self.__value
        if len(value.split('.')) >= 3:
            return value.split('.')[2]

    def get_patch(self):
        """Get patch version, if applicable"""
        value = self.__value
        if len(value.split('.')) >= 4:
            return value.split('.')[3]

    def has_major(self):
        """Return True if version has a major part"""
        if self.get_major():
            return True
        return False

    def has_minor(self):
        """Return True if version has a minor part"""
        if self.get_minor():
            return True
        return False

    def has_revision(self):
        """Return True if version has a revision part"""
        if self.get_revision():
            return True
        return False

    def has_patch(self):
        """Return True if version has a patch part"""
        if self.get_patch():
            return True
        return False

    def get_weight(self):
        """Return weight of the version"""
        return self.__weight

    def __eq__(self, other):
        """Compare two versions. Return 1 if the versions are the same, 0 otherwise"""
        if self.get_value() == other.get_value():  # compare name value (should be unique)
            return 1
        else:
            return 0



import click

from noora.plugins.mssql.update.UpdatePlugin import UpdatePlugin


@click.command()
@click.option('-v', '--version', required=False, help='The version to update to')
@click.option(
    '-h', '--host', required=True, default='localhost', help='The host to update the db on')
@click.option('-e', '--environment', required=False, help='The environment to update the db in')
@click.pass_obj
def cli(props, version, host, environment):
    """
    Update an MSSQL database to the specified version.
    """
    plugin = UpdatePlugin()
    plugin.execute(props, {'version': version, 'host': host, 'environment': environment})



import click

from noora.plugins.mysql.update.UpdatePlugin import UpdatePlugin


@click.command()
@click.option('-v', '--version', required=False)
@click.option('-h', '--host', required=True, default='localhost')
@click.option('-d', '--database', required=False, help='Database name')
@click.option('-e', '--environment', required=False, help='Environment')
@click.option('-a', '--alias', required=False, help='Alias. Overrules the database option')
@click.pass_obj
def cli(props, version, host, database, environment, alias):
    """
    Update a MSSQL database to the specified version.
    """
    plugin = UpdatePlugin()
    plugin.execute(
        props,
        {
            'version': version, 'host': host, 'database': database,
            'environment': environment, 'alias': alias
        },
    )



import os
import json


# FIXME: write a proper docstring
def get_database_folder(database, folder_aliases):
    if folder_aliases:
        for folder_alias in folder_aliases:
            alias = folder_alias[0]
            if database == alias:
                return folder_alias[1]
    return database


def get_mssql_properties(users, host, database, schema):
    """
    For a list of user tuples, find the first occurrence that matches host and database.

    :param users: list containing a sequence of AT LEAST (host, schema, username, password)
    :param database : the name of override database
    :param host: the hostname to match
    :param schema: the schema to match
    :return: a dict containing host, schema, username, password, port
    """
    for user in users:
        if user.get("host").lower() == host.lower() \
                and user.get("schema").lower() == schema.lower():
            res = {
                'host': host,
                'database': user.get("database", database),
                'schema': schema,
                'port': user.get("port"),
                'username': user.get("username"),
                'password': user.get("password"),
            }
            return res

    return None


def get_mysql_properties(users, host, database):
    """
    For a list of user tuples, find the first occurrence that matches host and database.

    :param users: list containing a sequence of AT LEAST (host, database, username, password)
    :param host: the hostname to match
    :param database: the database to match
    :return: a dict containing host, database, username, password, port
    """
    for user in users:
        if user.get("host").lower() == host.lower() \
                and user.get("database").lower() == database.lower():
            res = {
                'host': host,
                'database': database,
                'port': user.get("port"),
                'username': user.get("username"),
                'password': user.get("password"),
            }
            return res

    return None


def get_postgres_properties(users, host, database):
    """
    For a list of user tuples, return the first user that matches the host and database

    :param users: list containing a sequence of AT LEAST (host, database, username, password, port)
    :param host: The hostname to match
    :param database: The database to match
    :return: a dict containing host, database, username, password, port
    """
    for user in users:
        if user.get("host").lower() == host.lower() \
                and user.get("database").lower() == database.lower():
            res = {
                'host': host,
                'database': database,
                'port': user.get("port"),
                'username': user.get("username"),
                'password': user.get("password"),
            }
            return res

    return None


def get_oracle_user(users, sid, scheme):
    """
    For a list of sid/scheme/user combinations, return the username that matches the
    sid, scheme and user.

    :param users: list containing a sequence of AT LEAST (sid, scheme, username)
    :param sid: The Site Identifier (SID) to match
    :param scheme: The scheme to match
    :return: A username or None
    """
    for user in users:
        if user[0].lower() == sid.lower() and user[1].lower() == scheme.lower():
            return user[2]
    return None


def get_oracle_passwd(users, sid, scheme):
    """
    For a list of sid/scheme/user/password combinations, return the password that matches the
    sid, scheme and user.

    :param users: list containing a sequence of AT LEAST (sid, scheme, username, password)
    :param sid: The Site Identifier (SID) to match
    :param scheme: The scheme to match
    :return: A password or None
    """
    for user in users:
        if user[0].lower() == sid.lower() and user[1].lower() == scheme.lower():
            return user[3]
    return None


def get_profile(properties):
    """
    Retrieve the profile for the database project in the properties.
    """
    profile = None
    project = properties.get('project')
    if project:
        # load the users from the credentials file, if present
        home_dir = properties.get('home.dir')
        credentials_file = os.path.join(home_dir, '.noora', 'credentials.json')
        if os.path.isfile(credentials_file):
            f = open(credentials_file, 'r')
            credentials = json.load(f)
            f.close()

            profile = credentials.get(project)
    return profile


def connection_credentials(connection_string: str) -> dict:
    """
    Returns a dict containing the connection credentials
    The following format is expected for the connection_string.
    In the example below a postgresql connection string is used.
    --connection-string host=localhost,port=5432,database=acme,username=acme_owner,password=acme_owner

    :type connection_string: str
    :param connection_string: A connection string;
    """

    key_strings = connection_string.split(",")
    result = {}
    for key_string in key_strings:
        key_value_pair = key_string.split("=")
        key = key_value_pair[0]
        value = key_value_pair[1]
        result[key] = value
    return [
        result
    ]



from noora.exceptions.ConnectorException import ConnectorException


class Connectable(object):
    """
    Abstract connector class indicating if an object is 'connectable', i.e.,
    can connect to a database and execute queries.
    """
    def execute(self, executable, properties):
        raise ConnectorException("method not implemented")



from setuptools import setup, find_packages
import codecs
import os
import re

HERE = os.path.abspath(os.path.dirname(__file__))
META_PATH = os.path.join('noora', '__init__.py')


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string.'.format(meta=meta))


setup(
    name=find_meta('title'),
    version=find_meta('version'),
    description=find_meta('description'),
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license=find_meta('license'),
    author=find_meta('author'),
    author_email=find_meta('email'),
    maintainer=find_meta('author'),
    maintainer_email=find_meta('email'),
    url=find_meta('uri'),
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Database :: Front-Ends',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    keywords='development database',
    packages=find_packages(exclude=['doc', 'docs', 'examples', 'releases', 'snippets', 'src', 'test', 'tests']),
    package_data={'noora': ['*.json',
                            'plugins/mysql/generate/templates/*.json',
                            'plugins/mysql/generate/templates/fct/*.sql',
                            'plugins/mysql/generate/templates/idx/*.sql',
                            'plugins/mysql/generate/templates/tab/*.sql',
                            'plugins/mysql/generate/templates/trg/*.sql',
                            'plugins/mysql/drop/fct/*.sql',
                            'plugins/mysql/drop/prc/*.sql',
                            'plugins/mysql/drop/tab/*.sql',
                            'plugins/mysql/drop/vw/*.sql',
                            'plugins/mysql/update/*.sql',
                            'plugins/mssql/generate/templates/*.json',
                            'plugins/mssql/generate/templates/fct/*.sql',
                            'plugins/mssql/generate/templates/idx/*.sql',
                            'plugins/mssql/generate/templates/seq/*.sql',
                            'plugins/mssql/generate/templates/tab/*.sql',
                            'plugins/mssql/generate/templates/trg/*.sql',
                            'plugins/mssql/drop/cst/*.sql',
                            'plugins/mssql/drop/fct/*.sql',
                            'plugins/mssql/drop/prc/*.sql',
                            'plugins/mssql/drop/seq/*.sql',
                            'plugins/mssql/drop/syn/*.sql',
                            'plugins/mssql/drop/tab/*.sql',
                            'plugins/mssql/drop/typ/*.sql',
                            'plugins/mssql/drop/vw/*.sql',
                            'plugins/mssql/update/*.sql',
                            'plugins/postgresql/generate/templates/*.json',
                            'plugins/postgresql/generate/templates/cst/*.sql',
                            'plugins/postgresql/generate/templates/fct/*.sql',
                            'plugins/postgresql/generate/templates/idx/*.sql',
                            'plugins/postgresql/generate/templates/seq/*.sql',
                            'plugins/postgresql/generate/templates/tab/*.sql',
                            'plugins/postgresql/generate/templates/trg/*.sql',
                            'plugins/postgresql/drop/cst/*.sql',
                            'plugins/postgresql/drop/fct/*.sql',
                            'plugins/postgresql/drop/idx/*.sql',
                            'plugins/postgresql/drop/seq/*.sql',
                            'plugins/postgresql/drop/tab/*.sql',
                            'plugins/postgresql/drop/trg/*.sql',
                            'plugins/postgresql/drop/vw/*.sql',
                            'plugins/postgresql/update/*.sql']},
    install_requires=['six', 'click', 'fire'],
    entry_points={
        'console_scripts': [
            'mynoora=noora.mynoora_cli:main',
        ],
    },
)



import click

from noora.plugins.mssql.drop.DropPlugin import DropPlugin


@click.command()
@click.option('-h', '--host', required=True, default='localhost', help='The host to drop the database on')
@click.option('-e', '--environment', required=False, help='The environment to drop the database in')
@click.pass_obj
def cli(props, host, environment):
    """
    Drop (empty) an MSSQL database.
    """
    plugin = DropPlugin()
    plugin.execute(props, {'host': host, 'environment': environment})



#!/usr/bin/env python
import click

import noora

from noora.system.App import App


@click.command(cls=App, invoke_without_command=True, no_args_is_help=True)
@click.option('-r', '--revision', is_flag=True, help="Report the version of Noora and exit.")
@click.pass_context
def main(ctx, revision):
    """
    NoOra is a database deployment tool used to automate the database deployment cycle.
    """
    # show the revision
    if revision:
        print(noora.__title__ + " version " + noora.__version__)
        ctx.exit()


if __name__ == "__main__":
    main()



from noora.version.Version import Version


class Versions(object):
    """
    Management class for handling versions of a project. Useful for inspecting
    and testing the versions on a project and to manage updating projects to
    specific versions.
    """
    def __init__(self):
        self.__versions = []

    def clear(self):
        """Remove all versions from the class"""
        self.__versions = []

    def sort(self):
        """Sort the versions by weight"""
        self.__versions.sort(key=lambda version: version.get_weight(), reverse=False)

    def add(self, version):
        """Add a version to the list"""
        versions = self.__versions
        versions.append(version)

    def exists(self, other):
        """Check if the specified version exists"""
        versions = self.__versions
        for version in versions:
            if version == other:
                return True
        return False

    def list(self):
        """Return the list of versions"""
        return self.__versions

    def previous(self, other):
        """Return the version that precedes the provided version"""
        versions = self.__versions
        i = 0
        for version in versions:
            if version == other:
                if i == 0:
                    return version
                return versions[i - 1]
            i = i + 1

    def get_part(self, version):
        """Return the number of parts in the specified version"""
        result = 0
        if version.has_major():
            result = result + 1
        if version.has_minor():
            result = result + 1
        if version.has_revision():
            result = result + 1
        if version.has_patch():
            result = result + 1
        return result

    def next(self):
        """
        Calculate the next version number based on the last version in the
        list, incrementing the smallest version component by one depending on the
        number of parts.
        """
        last = self.last()
        if last:
            level = self.get_part(last)
            major = last.get_major()
            minor = last.get_minor()
            revision = last.get_revision()
            patch = last.get_patch()
            if level == 1:
                major = str(int(major) + 1)
            if level == 2:
                minor = str(int(minor) + 1)
            if level == 3:
                revision = str(int(revision) + 1)
            if level == 4:
                patch = str(int(patch) + 1)

            if level == 1:
                version = major
            elif level == 2:
                version = major + "." + minor
            elif level == 3:
                version = major + "." + minor + "." + revision
            elif level == 4:
                version = major + "." + minor + "." + revision + "." + patch
            # Higher levels are not supported
            else:
                # FIXME: raise exception?
                return None

            return Version(version)

        # If no version was found, return nothing.
        return None

    def last(self):
        """Return the last version in the list"""
        versions = self.__versions
        if versions:
            return versions[-1]



__version__ = '1.2.4'

__title__ = 'noora'
__description__ = 'NoOra a database deployment tool.'
__uri__ = 'https://github.com/janripke/noora'
__doc__ = __description__ + ' <' + __uri__ + '>'

__author__ = 'Jan Ripke'
__email__ = 'janripke@gmail.com'

__license__ = 'Apache License 2.0'
__copyright__ = 'Copyleft (l) 2021 Jan Ripke'




import os

from noora.version.Version import Version
from noora.version.Versions import Versions
from noora.version.VersionLoader import VersionLoader

from noora.system import Ora
from noora.system import PropertyHelper
from noora.io.File import File

from noora.plugins.postgresql.PGSQLPlugin import PGSQLPlugin
from noora.plugins.Fail import Fail
from noora.exceptions.plugins.InvalidEnvironmentException import InvalidEnvironmentException
from noora.exceptions.plugins.InvalidVersionException import InvalidVersionException

from noora.connectors.ConnectionExecutor import ConnectionExecutor


class UpdatePlugin(PGSQLPlugin):
    """This class provides functionality for updating a project to a specified version."""
    def _validate_and_prepare(self, properties, arguments):
        prepared_args = {}

        version = arguments.get('version')
        Fail.fail_on_no_version(version)
        Fail.fail_on_unknown_version(version, properties)
        prepared_args['version'] = version

        host = arguments.get('host')
        Fail.fail_on_no_host(host)
        prepared_args['host'] = host

        environment = arguments.get('environment')
        default_environment = properties.get('default_environment')
        environment = Ora.nvl(environment, default_environment)
        Fail.fail_on_invalid_environment(environment, properties)
        prepared_args['environment'] = environment

        prepared_args['connection_string'] = arguments.get('connection_string')

        alias = arguments.get('alias')
        Fail.fail_on_invalid_alias(alias, properties)
        # if an alias is given, only this database will be installed, other databases will be
        # ignored.
        if alias:
            print("using alias: {}".format(alias))
            prepared_args['databases'] = [alias]
        else:
            database = arguments.get('database')
            Fail.fail_on_invalid_database(database, properties)
            default_databases = properties.get('databases')
            databases = Ora.nvl(database, default_databases)
            prepared_args['databases'] = databases

        return prepared_args

    def fail_on_invalid_environment(self, connector, executor, environment, properties):
        plugin_dir = properties.get('plugin.dir')
        properties['environment'] = environment
        script = File(os.path.join(plugin_dir, 'postgresql', 'update', 'checkenvironment.sql'))
        executor['script'] = script
        connector.execute(executor, properties)
        if "(Code 1329)" in connector.get_result():
            raise InvalidEnvironmentException("invalid environment: {}".format(environment))

    def fail_on_invalid_version(self, connector, executor, version, properties):
        plugin_dir = properties.get('plugin.dir')

        versions = Versions()
        version_loader = VersionLoader(versions)
        version_loader.load(properties)
        versions.sort()
        v = Version(version)
        previous = versions.previous(v).get_value()

        properties['previous'] = previous
        script = File(os.path.join(plugin_dir, 'postgresql', 'update', 'checkversion.sql'))
        executor['script'] = script
        connector.execute(executor, properties)
        if "(Code 1329)" in connector.get_result():
            raise InvalidVersionException("invalid version: {}".format(previous))

    def execute(self, properties, arguments):
        """
        Update database after checking if schema and environment are
        valid values. Also check that host is not on the block list and that
        the version to update to is valid

        :type properties: system.Properties.Properties
        :param properties: The project properties
        :type arguments: dict
        :param arguments: This dict contains the plugin arguments:

            * **version**: The version to update the database to;
            * **host**: The hostname that hosts the database to update;
            * **schema**: Schema to update (optional);
            * **environment**: Environment to update the database in (optional).
        """
        prepared_args = self._validate_and_prepare(properties, arguments)

        version = prepared_args['version']
        host = prepared_args['host']
        environment = prepared_args['environment']
        databases = properties.get('databases')

        objects = properties.get('create_objects')

        version_database = properties.get('version_database')

        alter_dir = properties.get('alter.dir')

        # retrieve the database connection credentials through
        # the commandline option --connection-string
        connection_string = prepared_args['connection_string']
        if connection_string:
            users = PropertyHelper.connection_credentials(connection_string)

        if not connection_string:
            # retrieve the user credentials for this database project.
            users = properties.get('postgresql_users')

            # try to retrieve the users from the credentials file, when no users are configured in
            # myproject.json.
            if not users:
                # retrieve the name of this database project, introduced in version 1.0.12
                profile = PropertyHelper.get_profile(properties)
                if profile:
                    users = profile.get('postgresql_users')

        connector = self.get_connector()

        for database in databases:
            print("updating database '{db}' on host '{host}' using environment '{env}'".format(
                db=database, host=host, env=environment))

            executor = PropertyHelper.get_mysql_properties(users, host, database)

            if database == version_database:
                self.fail_on_invalid_environment(connector, executor, environment, properties)
                self.fail_on_invalid_version(connector, executor, version, properties)

            for obj in objects:
                # global ddl objects
                folder = File(os.path.join(alter_dir, version, database, 'ddl', obj))
                ConnectionExecutor.execute(connector, executor, properties, folder)

                # environment specific ddl objects
                folder = File(
                    os.path.join(alter_dir, version, database, 'ddl', obj, environment))
                ConnectionExecutor.execute(connector, executor, properties, folder)

            # global dat objects
            folder = File(os.path.join(alter_dir, version, database, 'dat'))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            # environment specific dat objects
            folder = File(os.path.join(alter_dir, version, database, 'dat', environment))
            ConnectionExecutor.execute(connector, executor, properties, folder)

            print("database '{}' updated".format(database))
